import cv2
import numpy as np

from gbvision.utils.pipeline import PipeLine


def _corners(im):
    return cv2.filter2D(im, -1, np.array([[-1, 1], [1, -1]]))


corners = PipeLine(_corners)


def _edges(im):
    return cv2.filter2D(im, -1, np.array([[-1, -1, -1], [-1, 8, -1], [-1, -1, -1]]))


edges = PipeLine(_edges)


def _sharpen(im):
    return cv2.filter2D(im, -1, np.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]]))


sharpen = PipeLine(_sharpen)


def _blur(im):
    return cv2.filter2D(im, -1, np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]) / 9)


blur = PipeLine(_blur)


def _blue(im):
    return im[:, :, 0]


blue = PipeLine(_blue)


def _green(im):
    return im[:, :, 1]


green = PipeLine(_green)


def _red(im):
    return im[:, :, 2]


red = PipeLine(_red)


def _gray(im):
    return cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)


gray = PipeLine(_gray)





from threading import Thread

import cv2

from gbvision.net.stream_receiver import StreamReceiver
from gbvision.constants.system import EMPTY_PIPELINE
from .window import Window


class StreamWindow(Window):
    """
    a basic window that displays the stream from a stream receiver
    """

    def __init__(self, stream_receiver: StreamReceiver, window_name='stream', exit_button='qQ',
                 drawing_pipeline=EMPTY_PIPELINE):
        """
        initializes the stream window
        :param stream_receiver: a stream receiver from which the stream receives it's frames
        :param drawing_pipeline: optional, a pipeline of drawing functions that will run on the frame before displaying
        it
        :param exit_button: an array of keys (a string), when one of the keys are pressed the window will be closed
        :param window_name: the title of the window
        """
        Window.__init__(self, window_name, exit_button, drawing_pipeline)
        self.stream_receiver = stream_receiver

    def show(self, flags=cv2.WINDOW_FREERATIO):
        """
        display the stream video window
        :param flags: some opencv window flags
        """
        cv2.namedWindow(self.window_name, flags)
        while True:
            frame = self.stream_receiver.get_frame()
            if frame is not None:
                cv2.imshow(self.window_name, self.drawing_pipeline(frame))
            k = chr(cv2.waitKey(1) & 0xFF)
            if k in self.exit_button:
                cv2.destroyWindow(self.window_name)
                return

    def show_async(self, flags=cv2.WINDOW_FREERATIO):
        """
        opens the steam video window on another thread
        :param flags: some opencv window flags
        """
        Thread(target=self.show, args=(flags,)).start()



from gbvision.utils.camera_data import CameraData

LIFECAM_STUDIO = CameraData(648.5256168410046, 0.340394)

LIFECAM_3000 = CameraData(697.0395744431028, 0.4228539261329407)

UNKNOWN_CAMERA = CameraData(0, 0)



import numpy as np

from gbvision.models.contours import filter_contours, find_contours, sort_contours, contours_to_rotated_rects, \
    contours_to_polygons
from gbvision.tools.list_tools import split_list
from .object_finder import ObjectFinder
from gbvision.constants.system import EMPTY_PIPELINE


class TargetPairFinder(ObjectFinder):
    """
    finds a pair of vision targets
    """

    def __init__(self, threshold_func, game_object, vt_distance=0.2866, enclosing_rect_ratio=0.549,
                 contour_min_area=3.0):
        """
        initializes the finder
        :param vt_distance: the distance between the centers of both vision targets
        :param enclosing_rect_ratio: the ratio between the width and height of the parallel enclosing rect of the vision
        target
        :param contour_min_area: the minimal area of a contour, used in filter_contours
        """
        ObjectFinder.__init__(self, threshold_func, game_object)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               filter_contours(min_area=contour_min_area) +
                               sort_contours)
        self.__vector_distance = np.array([vt_distance / 2, 0, 0])
        self.vt_distance = vt_distance
        self.enclosing_rect_ratio = enclosing_rect_ratio

    def __call__(self, frame, camera):
        cnts = self._full_pipeline(frame)
        rects = contours_to_rotated_rects(cnts)
        polys = contours_to_polygons(cnts)
        rects_polys = zip(rects, polys)
        left_targets_polys, right_targets_polys = split_list(
            lambda rotated_rect: rotated_rect[0][2] < -45.0, rects_polys)

        left_targets = list(map(lambda pair: pair[0], left_targets_polys))
        right_targets = list(map(lambda pair: pair[0], right_targets_polys))

        left_targets_real, right_targets_real = [], []
        for i in left_targets:
            left_targets_real.append(self.game_object.location3d_by_params(camera, np.sqrt(i[1][0] * i[1][1]), i[0]))
        for i in right_targets:
            right_targets_real.append(self.game_object.location3d_by_params(camera, np.sqrt(i[1][0] * i[1][1]), i[0]))

        target_pairs = []
        i = 0
        while i < len(left_targets_real):
            lt = left_targets_real[i]
            possibles = sorted(filter(lambda t: abs(np.linalg.norm(lt - t[1]) - self.vt_distance) < 0.2,
                                      enumerate(right_targets_real)),
                               key=lambda t: abs(np.linalg.norm(lt - t[1]) - self.vt_distance))
            for p in possibles:
                if right_targets[p[0]][0][0] < left_targets[i][0][0]:
                    target_pairs.append((lt, p[1]))
                    del left_targets[i]
                    del left_targets_real[i]
                    del left_targets_polys[i]
                    del right_targets[p[0]]
                    del right_targets_real[p[0]]
                    del right_targets_polys[p[0]]
                    i -= 1
                    break
            i += 1
        all_hatches = []
        for i in target_pairs:
            all_hatches.append(
                np.concatenate(((i[0] + i[1]) / 2,
                                np.array(
                                    [np.pi / 2 - np.arccos(
                                        max(-1, min(1, (i[0][2] - i[1][2]) / (self.vt_distance * 2))))]))))

        for i, t in enumerate(left_targets_real):
            if len(left_targets_polys[i][1]) != 4:
                print('[WARN] polydp returning %d points instead of four' % len(left_targets[i][1]))
                continue

            width, height = left_targets[i][1]
            tmp_ang = np.deg2rad(90.0 + left_targets[i][2])

            w_s = np.cos(tmp_ang) * width + np.sin(tmp_ang) * height
            h_s = np.sin(tmp_ang) * width + np.cos(tmp_ang) * height

            poly = np.array(left_targets_polys[i][1])

            sorted_polys = sorted(poly, key=lambda p: p[1])
            highest = sorted_polys[0]
            lowest = sorted_polys[3]

            sorted_polys = sorted(poly, key=lambda p: p[0])
            leftest = sorted_polys[0]
            rightest = sorted_polys[3]

            sng = np.sign(np.linalg.norm(rightest - highest) - np.linalg.norm(leftest - lowest))
            angle = sng * np.arccos(min(min(w_s / h_s, h_s / w_s) / self.enclosing_rect_ratio, 1))

            rot_matrix = np.array([[np.cos(angle), 0, np.sin(angle)],
                                   [0, 1, 0],
                                   [-np.sin(angle), 0, np.cos(angle)]])
            all_hatches.append(np.concatenate((t - rot_matrix.dot(self.__vector_distance), np.array([-angle]))))

        for i, t in enumerate(right_targets_real):
            if len(right_targets_polys[i][1]) != 4:
                print('[WARN] polydp returning %d points instead of four' % len(right_targets[i][1]))
                continue
            width, height = right_targets[i][1]
            tmp_ang = np.deg2rad(abs(right_targets[i][2]))

            w_s = np.cos(tmp_ang) * width + np.sin(tmp_ang) * height
            h_s = np.sin(tmp_ang) * width + np.cos(tmp_ang) * height

            poly = np.array(right_targets_polys[i][1])

            sorted_polys = sorted(poly, key=lambda p: p[1])
            highest = sorted_polys[0]
            lowest = sorted_polys[3]

            sorted_polys = sorted(poly, key=lambda p: p[0])
            leftest = sorted_polys[0]
            rightest = sorted_polys[3]

            sng = np.sign(np.linalg.norm(rightest - lowest) - np.linalg.norm(leftest - highest))
            angle = sng * np.arccos(min(min(w_s / h_s, h_s / w_s) / self.enclosing_rect_ratio, 1))

            rot_matrix = np.array([[np.cos(angle), 0, np.sin(angle)],
                                   [0, 1, 0],
                                   [-np.sin(angle), 0, np.cos(angle)]])
            all_hatches.append(np.concatenate((t + rot_matrix.dot(self.__vector_distance), np.array([-angle]))))
        all_hatches.sort(key=lambda v: np.linalg.norm(v[0:3:2]), reverse=False)
        return all_hatches



import numpy as np

from gbvision.exceptions import AbstractMethodCallingException


class StreamReceiver:
    """
    this is an abstract receiver that receives stream from a broadcast receiver
    this class should not be instanced but inherited from
    """

    def __init__(self, fx: float = 1.0, fy: float = 1.0):
        """
        initializes a stream receiver
        :param fx: ratio between width of the read frame to the width of the frame returned, default is 1 (same width)
        :param fy: ratio between height of the read frame to the height of the frame returned,
         default is 1 (same height)
        """
        self.fx = fx
        self.fy = fy

    def get_frame(self) -> np.ndarray:
        """
        reads a frame from the stream and returns in
        :returns: the frame read as a numpy array
        """
        raise AbstractMethodCallingException()



import pickle
import socket
import struct
import time

import cv2

from gbvision.constants.net import LOCAL_SERVER_IP
from .stream_broadcaster import StreamBroadcaster


class TCPStreamBroadcaster(StreamBroadcaster):
    """
    this class uses TCP to send a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the server and the receiver is the client
    """

    def __init__(self, port: int, fx: float = 1.0, fy: float = 1.0, im_encode: str = '.jpg',
                 use_grayscale: bool = False,
                 max_fps: int = None):
        """
        initializes the streamer
        :param port: the port which TCP will be using
        :param im_encode: the type of image encoding to send over the network, default is .jpg (JPEG)
        for missing parameters, see documentation on StreamBroadcaster
        """
        StreamBroadcaster.__init__(self, fx=fx, fy=fy, use_grayscale=use_grayscale, max_fps=max_fps)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_addr = (LOCAL_SERVER_IP, port)
        self.socket.bind(self.server_addr)
        self.socket.listen(10)
        self.socket, addr = self.socket.accept()
        self.payload_size = struct.calcsize("I")
        self.im_encode = im_encode
        self.prev_time = 0.0

    def send_frame(self, frame):
        if self.max_fps is not None and (time.time() - self.prev_time) * self.max_fps < 1:
            return
        if frame is not None:
            frame = cv2.resize(frame, (0, 0), fx=self.fx, fy=self.fy)
            if self.use_grayscale:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

            frame = cv2.imencode(self.im_encode, frame)[1]
        data = pickle.dumps(frame)
        data = struct.pack("I", len(data)) + data
        self.socket.send(data)  # to(data, self.server_addr)
        self.prev_time = time.time()



from .object_finder import ObjectFinder
from .circle_finder import CircleFinder
from .rect_finder import RectFinder
from .polygon_finder import PolygonFinder
from .rotated_rect_finder import RotatedRectFinder
from .target_pair_finder import TargetPairFinder




import cv2
import numpy as np

from gbvision.models.contours import filter_contours, find_contours, sort_contours, contour_center
from gbvision.constants.system import EMPTY_PIPELINE
from .object_finder import ObjectFinder


class PolygonFinder(ObjectFinder):
    """
    finds any generic polygon, not recommended when another finder can be used
    """

    def __init__(self, threshold_func, game_object, area_scalar=1.0, contour_min_area=3.0):
        """

        :param area_scalar: optional, a scalar to multiply the area by, for fine tuning of the function's output
        :param contour_min_area:  the minimal area of a contour, used in filter_contours
        """
        ObjectFinder.__init__(self, threshold_func, game_object)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               filter_contours(min_area=contour_min_area) +
                               sort_contours)
        self.area_scalar = area_scalar

    def __call__(self, frame, camera):
        contours = self._full_pipeline(frame)
        return list(map(
            lambda cnt: self.game_object.location3d_by_params(camera, self.area_scalar * np.sqrt(cv2.contourArea(cnt)),
                                                              contour_center(cnt)), contours))



from .vision_warning import VisionWarning
from .vision_exception import VisionException
from .could_not_read_frame_exception import CouldNotReadFrameException
from .device_not_found_exception import DeviceNotFoundException
from .abstract_method_calling_exception import AbstractMethodCallingException



from .object_finder import ObjectFinder
from gbvision.models.contours import find_contours, filter_contours, sort_contours, contours_to_rects_sorted
from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.models.shapes import filter_inner_rects
import numpy as np


class RectFinder(ObjectFinder):
    """
    finds a rectangular shaped object
    """
    def __init__(self, threshold_func, game_object, area_scalar=1.0, contour_min_area=3.0):
        """

        :param area_scalar: optional, a scalar to multiply the area by, for fine tuning of the function's output
        :param contour_min_area: the minimal area of a contour, used in filter_contours
        """
        ObjectFinder.__init__(self, threshold_func, game_object)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               filter_contours(min_area=contour_min_area) +
                               sort_contours +
                               contours_to_rects_sorted +
                               filter_inner_rects)
        self.area_scalar = area_scalar

    def __call__(self, frame, camera):
        rects = self._full_pipeline(frame)
        return list(map(
            lambda rect: self.game_object.location3d_by_params(camera, self.area_scalar * np.sqrt(rect[2] * rect[3]),
                                                             [(rect[0] + rect[2]) / 2, (rect[1] + rect[3]) / 2]), rects))
        # d = []
        # for rect in rects:
        #    area = self.area_scalar * np.sqrt(rect[2] * rect[3])
        #    center = [(rect[0] + rect[2]) / 2, (rect[1] + rect[3]) / 2]
        #    d.append(self.im_object.location3d_by_params(camera, area, center))
        # return d



from .camera import CameraData, Camera
import cv2


class USBCamera(cv2.VideoCapture, Camera):
    """
    a basic usb connected camera which inherits from cv2 VideoCapture
    """
    def __init__(self, port, data: CameraData):
        """
        initializes the camera
        :param port: the usb port to which the camera is connected
        :param data: the camera data object that describes this camera
        """
        cv2.VideoCapture.__init__(self, port)
        self._data = data
        self.port = port

    def is_opened(self) -> bool:
        return self.isOpened()

    def set_exposure(self, exposure: int or float or bool) -> bool:
        if type(exposure) is bool:
            return self.set(cv2.CAP_PROP_EXPOSURE, int(exposure))
        return self.set(cv2.CAP_PROP_EXPOSURE, exposure)

    def set_auto_exposure(self, auto: int or float or bool) -> bool:
        if type(auto) is bool:
            return self.set(cv2.CAP_PROP_AUTO_EXPOSURE, 0.75 if auto else 0.25)
        return self.set(cv2.CAP_PROP_AUTO_EXPOSURE, auto)

    @property
    def data(self):
        return self._data

    @property
    def width(self):
        return self.get(cv2.CAP_PROP_FRAME_WIDTH)

    @property
    def height(self):
        return self.get(cv2.CAP_PROP_FRAME_HEIGHT)

    def _set_width(self, width):
        return self.set(cv2.CAP_PROP_FRAME_WIDTH, width)

    def _set_height(self, height):
        return self.set(cv2.CAP_PROP_FRAME_HEIGHT, height)



import cv2
import numpy as np

from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.models.contours import find_contours, contours_to_circles, contours_to_rects, contours_to_rotated_rects,\
    contours_to_ellipses
from gbvision.utils.pipeline import PipeLine


def draw_contours(threshold_func, color: (int, int, int), thickness=2, contours_process=EMPTY_PIPELINE):
    contour_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process

    def _draw(frame):
        frame = frame.copy()
        cnts = contour_finding(frame)
        cv2.drawContours(frame, cnts, -1, color, thickness)
        return frame

    return PipeLine(_draw)


def draw_circles(threshold_func, color: (int, int, int), thickness=2, contours_process=EMPTY_PIPELINE,
                 circle_process=EMPTY_PIPELINE):
    circle_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_circles + \
                     circle_process

    def _draw(frame):
        frame = frame.copy()
        circs = circle_finding(frame)
        for c in circs:
            cv2.circle(frame, (int(c[0][0]), int(c[0][1])), int(c[1]), color, thickness)
        return frame

    return PipeLine(_draw)


def draw_rects(threshold_func, color: (int, int, int), thickness=2, contours_process=EMPTY_PIPELINE,
               rects_process=EMPTY_PIPELINE):
    rect_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_rects + \
                   rects_process

    def _draw(frame):
        frame = frame.copy()
        rects = rect_finding(frame)
        for r in rects:
            cv2.rectangle(frame, (int(r[0]), int(r[1])), (int(r[0] + r[2]), int(r[1] + r[3])), color, thickness)
        return frame

    return PipeLine(_draw)


def draw_rotated_rects(threshold_func, color: (int, int, int), thickness=2, contours_process=EMPTY_PIPELINE,
                       rotated_rects_process=EMPTY_PIPELINE):
    rotated_rect_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_rotated_rects + \
                           rotated_rects_process

    def _draw(frame):
        frame = frame.copy()
        rotated_rects = rotated_rect_finding(frame)
        for r in rotated_rects:
            box = cv2.boxPoints(r)
            box = np.int0(box)
            cv2.drawContours(frame, [box], 0, color, thickness)
        return frame

    return PipeLine(_draw)


def draw_ellipses(threshold_func, color: (int, int, int), thickness=2, contours_process=EMPTY_PIPELINE,
                  ellipses_process=EMPTY_PIPELINE):
    ellipses_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_ellipses + \
                           ellipses_process

    def _draw(frame):
        frame = frame.copy()
        ellipses = ellipses_finding(frame)
        for e in ellipses:
            cv2.ellipse(frame, e, color, thickness)
        return frame

    return PipeLine(_draw)



import cv2
import numpy as np

from gbvision.utils.pipeline import PipeLine


def erode(ksize: int or (int, int), iterations=1):
    """
    creates a pipeline that erodes the image by a kernel of ones
    used mainly for erode & dilate denoise filters
    :param ksize: the kernel size, either an integer (meaning an nxn kernel) or a tuple (nxm kernel)
    :param iterations: optional, the amount of erode iterations to perform, default is 1
    None! a large number of iterations will slow down the program
    :return: a pipeline that erodes the given frame
    """
    if type(ksize) is int:
        ksize = (ksize, ksize)
    return PipeLine(lambda frame: cv2.erode(frame, np.ones(ksize), iterations=iterations))


def dilate(ksize: int or (int, int), iterations=1):
    """
    creates a pipeline that dilates the image by a kernel of ones
    used mainly for erode & dilate denoise filters
    :param ksize: the kernel size, either an integer (meaning an nxn kernel) or a tuple (nxm kernel)
    :param iterations: optional, the amount of dilate iterations to perform, default is 1
    None! a large number of iterations will slow down the program
    :return: a pipeline that dilates the given frame
    """
    if type(ksize) is int:
        ksize = (ksize, ksize)
    return PipeLine(lambda frame: cv2.dilate(frame, np.ones(ksize), iterations=iterations))


def median_blur(ksize: int):
    """
    creates a pipeline that blurs the given frame using the median blur method
    works very good for denoising purposes
    :param ksize: the size of the kernel used by the filter, must be an odd number
    :return: a pipeline that filters the image using the median blur method
    """
    return PipeLine(lambda frame: cv2.medianBlur(frame, ksize))



from .window import Window
from .stream_window import StreamWindow
from .feed_window import FeedWindow
from .camera_window import CameraWindow
from .drawing_tools import draw_circles, draw_contours, draw_ellipses, draw_rects, draw_rotated_rects



from os.path import splitext
from threading import Thread

import cv2

from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.constants.video import VIDEO_FILE_TYPE
from gbvision.utils.camera import Camera
from .window import Window


class RecordingCameraWindow(Window):
    """
    a basic window that displays the stream from a stream receiver
    """

    def __init__(self, camera: Camera, file_name: str, window_name='stream', fps=20.0, exit_button='qQ',
                 drawing_pipeline=EMPTY_PIPELINE):
        """
        initializes the stream window
        :param camera: the camera to display the window from
        :param file_name: the name of the output file
        :param drawing_pipeline: optional, a pipeline of drawing functions that will run on the frame before displaying
        it
        :param exit_button: an array of keys (a string), when one of the keys are pressed the window will be closed
        :param window_name: the title of the window
        :param fps: the fps of the video file
        """
        Window.__init__(self, window_name, exit_button, drawing_pipeline)
        self.camera = camera
        self.file_name = file_name

        _, file_ext = splitext(file_name)

        self.fourcc = cv2.VideoWriter_fourcc(*VIDEO_FILE_TYPE[file_ext.upper()])

        self.fps = fps

    def show(self, flags=cv2.WINDOW_FREERATIO):
        """
        display the stream video window
        :param flags: some opencv window flags
        """
        cv2.namedWindow(self.window_name, flags)
        video_writer = cv2.VideoWriter(self.file_name, self.fourcc, self.fps,
                                       (int(self.camera.width), int(self.camera.height)))
        while True:
            ok, frame = self.camera.read()
            if ok:
                video_writer.write(frame)
                cv2.imshow(self.window_name, self.drawing_pipeline(frame))
            k = chr(cv2.waitKey(1) & 0xFF)
            if k in self.exit_button:
                cv2.destroyWindow(self.window_name)
                return

    def show_async(self, flags=cv2.WINDOW_FREERATIO):
        """
        opens the steam video window on another thread
        :param flags: some opencv window flags
        """
        Thread(target=self.show, args=(flags,)).start()



import sys
from functools import reduce

import cv2


class ThresholdGroup:
    """
    a class that constructs a threshold filter out of several threshold filters and a binary mask function
    for example, use of the function on two thresholds with the binary function "bitwise_or" will result
    in a filter that outputs 255 for a pixel if it is in either one of the threshold's range
    using the "bitwise_and" function will output 255 for a pixel only if it is in both the threshold's range
    """

    def __init__(self, *thresholds, **kwargs):
        """
        initializes the threshold group
        :param thresholds: all the thresholds to join in the threshold group
        :param binary_mask: a binary function that maps from a pair of binary images to a single binary image
        default value is cv2.bitwise_or
        :param default_pixel: the default value of a pixel before the threshold function
        when using the 'bitwise_or' function this should be initialized to 0, when using cv2.bitwise_and this should be
        initialized to 255
        """
        self.binary_mask = cv2.bitwise_or
        self.default_pixel = 0
        if 'binary_mask' in kwargs:
            self.binary_mask = kwargs['binary_mask']
            del kwargs['binary_mask']
        if 'default_pixel' in kwargs:
            self.default_pixel = kwargs['default_pixel']
            del kwargs['default_pixel']
        for i in kwargs:
            print('[WARN] keyword value %s is never used' % i, file=sys.stderr)
        self.thresholds = list(thresholds)

    def __call__(self, frame):
        """
        apply the threshold filter to the frame
        :param frame: the frame to apply the filter to
        :return: the binary image, the frame after the threshold group filter
        """
        return reduce(lambda th_frame, threshold: self.binary_mask(th_frame, threshold(frame)), self.thresholds,
                      self.default_pixel)

    def __add__(self, other):
        """
        adds another threshold to the group and returns the new threshold group
        :param other: a threshold function ,can be a threshold, threshold group, or any type of function
        :return: a new threshold group with the other threshold joined
        """
        return ThresholdGroup(self.thresholds + [other], binary_mask=self.binary_mask, default_pixel=self.default_pixel)

    def __iadd__(self, other):
        """
        adds a new threshold filter to this threshold group
        :param other: a threshold function ,can be a threshold, threshold group, or any type of function
        """
        self.thresholds += [other]

    def __iter__(self):
        """
        :return: an iterator that iterates through all this group's filters
        """
        return iter(self.thresholds)

    def __len__(self):
        """
        :return: the amount of filters in this threshold group
        """
        return len(self.thresholds)

    def __getitem__(self, item):
        return self.thresholds[item]

    def __setitem__(self, key, value):
        self.thresholds[key] = value

    def __or__(self, other):
        return ThresholdGroup(self, other, binary_mask=cv2.bitwise_or, default_pixel=0)

    def __and__(self, other):
        return ThresholdGroup(self, other, binary_mask=cv2.bitwise_and, default_pixel=0xFF)



class CameraData:
    """
    describes constant about a camera in it's default state used to approximate distance
    between the camera and an object seen in a frame
    """
    def __init__(self, focal_length, fov):
        """

        :param focal_length: the focal length of the camera at it's default state, in units of pixels
        can be described as the square root of the amount of pixels an object takes on a frame, multiplied by it's
        distance from the camera and divided by the square root of it's surface
        FOCAL_LENGTH = sqrt(P) * D / sqrt(W*H)
        where P is the amount of pixels in the frame representing the object,
        D is the real life distance between the object and the camera
        W is the real life width of the object
        H is the real life height of the object
        note that this is a constant, whatever object you choose to use, this formula will yield the same result
        :param fov:
        half the viewing angle of the camera (field of view) in radians, can be calculated by placing an object in front
        of the camera, so that the entire object is captured and it's center is at the frame's center.
        the tangent of the angle can be described as the width of the object in real life, divided by the
        product of the object's distance from the camera in real life and the ratio between the width of the frame
        in pixels and the width of the object in the frame, also in pixels
        tan(FOV) = (Wm) / (D * (Wp/Wf))
        where Wm is the real life width of the object
        D is the real life distance between the object and the camera
        Wp is the width of the object in the frame (pixels unit)
        Wf is the width of the frame (pixels unit)
        to calculate the FOV just apply the inverse tangent
        FOV = arctan(tan(FOV))
        """
        self.focal_length = focal_length
        self.fov = fov

    def __copy__(self):
        return CameraData(self.focal_length, self.fov)

    copy = __copy__



import cv2

from gbvision.utils.pipeline import PipeLine

CONTOURS_INDEX = 1 if cv2.getVersionString()[0] == '3' else 0

EMPTY_PIPELINE = PipeLine()


def cv_config():
    """
    configure some opencv stuff so that it doesn't cause problems
    """
    if cv2.__version__[0] == '2':
        for i in [attr for attr in dir(cv2.cv) if attr.startswith("CV_CAP_PROP")]:
            object.__setattr__(cv2, i[3:], cv2.cv.__getattribute__(i))
            cv2.__dict__[i[3:]] = cv2.cv.__dict__[i]



import cv2

from gbvision.constants.math import EPSILON
from gbvision.constants.system import CONTOURS_INDEX
from gbvision.utils.pipeline import PipeLine

find_contours = PipeLine(lambda frame: cv2.findContours(frame, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[CONTOURS_INDEX])

sort_contours = PipeLine(lambda cnts: sorted(cnts, key=lambda x: cv2.contourArea(x), reverse=True))


def filter_contours(min_area: float):
    return PipeLine((lambda cnts: filter(lambda c: cv2.contourArea(c) >= min_area, cnts))) + list


convex_hull = PipeLine(cv2.convexHull)

contour_center = PipeLine(lambda cnt: cv2.moments(cnt),
                          lambda m: (int(m['m10'] / (m['m00'] + EPSILON)), int(m['m01'] / (m['m00'] + EPSILON))))

contours_centers = PipeLine(lambda cnts: map(contour_center, cnts)) + list

# SHAPES

contours_to_rects = PipeLine(lambda cnts: map(cv2.boundingRect, cnts)) + list

sort_rects = PipeLine(lambda rects: sorted(rects, key=lambda x: x[2] * x[3], reverse=True)) + list

contours_to_rects_sorted = contours_to_rects + sort_rects

contours_to_circles = PipeLine(lambda cnts: map(cv2.minEnclosingCircle, cnts)) + list

sort_circles = PipeLine(lambda rects: sorted(rects, key=lambda x: x[1], reverse=True)) + list

contours_to_circles_sorted = contours_to_circles + sort_circles

contours_to_rotated_rects = PipeLine(lambda cnts: map(cv2.minAreaRect, cnts)) + list

sort_rotated_rects = PipeLine(lambda rects: sorted(rects, key=lambda x: x[1][0] * x[1][1], reverse=True)) + list

contours_to_rotated_rects_sorted = contours_to_rotated_rects + sort_rotated_rects

contours_to_ellipses = PipeLine(lambda cnts: filter(lambda x: len(x) >= 5, cnts),
                                # ellipse must get contours of at least five points
                                lambda cnts: map(cv2.fitEllipse, cnts)) + list

sort_ellipses = sort_rotated_rects

contours_to_ellipses_sorted = contours_to_ellipses + sort_ellipses

contours_to_polygons = PipeLine(lambda cnts: map(lambda cnt: (cnt, 0.05 * cv2.arcLength(cnt, True)), cnts),
                                lambda cnts: map(lambda cnt0_eps1: cv2.approxPolyDP(cnt0_eps1[0], cnt0_eps1[1], True),
                                                 cnts),
                                lambda polydps: map(lambda polydp: map(lambda x: x[0], polydp), polydps),
                                lambda polydps: map(lambda polydp: list(map(tuple, polydp)), polydps)) + list



from gbvision.constants.math import SQRT_PI
from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.models.contours import find_contours, filter_contours, contours_to_circles_sorted
from gbvision.models.shapes import filter_inner_circles
from .object_finder import ObjectFinder


class CircleFinder(ObjectFinder):
    """
    finds a circular shaped object, like a ball or a disk
    """
    def __init__(self, threshold_func, game_object, contour_min_area=3.0):
        """
        initializes the finder
        :param contour_min_area: the minimal area of a contour, used in filter_contours
        """
        ObjectFinder.__init__(self, threshold_func, game_object)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               filter_contours(min_area=contour_min_area) +
                               contours_to_circles_sorted +
                               filter_inner_circles)

    def __call__(self, frame, camera):
        circles = self._full_pipeline(frame)
        return list(map(lambda circ: self.game_object.location3d_by_params(camera, SQRT_PI * circ[1], circ[0]), circles))
